#!/usr/bin/perl -w

# pd.pl -- parses textual pd patches into proper pd patches

#    (c) yaxu 2008

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA


use strict;

my @lines;

while (<>) {
    chomp;
    push @lines, $_;
}

my @objects;
my $ids = 0;

foreach my $y (0 .. $#lines) {
    my $line = $lines[$y];
    my $obj = 0;
    my $n = 0;
    while($n < length($line)) {
	my $s = substr($line, $n);

	if ($s =~ m,(.*?)\|\s*([*/\+\-~\w]+)((?:\s+\d[\d\.]*)*)\s*\|,) {
	  my $len = length($&);
	  my $x = $n + length($1);
	  my $name = $2;
	  my $args = $3;
	  $args =~ s/^\s+//;
	  $args =~ s/\s+$//;
	  my @args = split(/\s+/, $args);

	  my $top = substr($lines[$y - 1], $x);
	  $top =~ s/(\.[^\.]+\.).*/$1/;
	  my $bot = substr($lines[$y + 1], $x);
	  $bot =~ s/(\.[^\.]+\.).*/$1/;

	  my @inlets = map {[$x + $_, $y - 1]} indexes($top, 'x');
	  my @outlets = map {[$x + $_, $y + 1]} indexes($bot, 'x');
	  

	  push(@objects, 
	       {id => $ids++,
		x => $x,
		y => $y - 1,
		name => $name,
		args => \@args,
		inlets => \@inlets,
		outlets => \@outlets,
	       }
	       
	      );
	  $n += $len;
	}
	else {
	  last;
	}
      }
}

my %inlets;
foreach my $object (@objects) {
  my $i = 0;
  foreach my $inlet (@{$object->{inlets}}) {
    $inlets{$inlet->[0] . '_' . $inlet->[1]} = [$object->{id}, $i];
    ++$i;
  }
}

print "#N canvas 0 0 450 300 10;\n";
foreach my $object (@objects) {
  my $args = '';
  if (@{$object->{args}}) {
    $args = ' ' . join(' ', @{$object->{args}});
  }
  my $x = $object->{x} * 10;
  my $y = $object->{y} * 10;
  print "#X obj $x $y $object->{name}$args;\n";
}

my @links;
foreach my $object (@objects) { 
  my $n = 0;
  foreach my $outlet (@{$object->{outlets}}) { 
    my @targets = trace(@$outlet, 'd');
    foreach my $target (@targets) {
      print "#X connect $object->{id} $n " . join(' ', @$target) . ";\n";
    }
    $n++;
  }
}

sub indexes {
  my ($str, $substr)  = @_;
  my @result;
  my $n = 0;
  while (($n = index($str, $substr, $n)) >= 0) {
    push(@result, $n);
    ++$n;
  }
  return(@result);
}

sub trace {
  my ($x, $y, $dir) = @_;
  my @result;
  if ($dir eq 'd') {
    if (at($x-1, $y+1) eq '/') {
      push(@result, trace($x-1, $y+1, [-1, 1]));
    }
    if (at($x, $y+1) eq '|') {
      push(@result, trace($x, $y+1, [0, 1]));
    }
    if (at($x+1, $y+1) eq '\\') {
      push(@result, trace($x+1, $y+1, [1, 1]));
    }
  }
  elsif ($dir eq '*') {
    my $predir = $_[3];
    my $tries = [
		 [-1, -1, '\\'],
		 [0,  -1, '|' ],
		 [1,  -1, '/' ],
		 [-1, 0,  '-' ],
		 [1,  0,  '-' ],
		 [-1, 1,  '/' ],
		 [0,  1,  '|' ],
		 [1,  1,  '\\'],
		];
    foreach my $try (@$tries) {
      next if (
	       ((0 - $try->[0]) == $predir->[0])
	       and
	       ((0 - $try->[1]) == $predir->[1])
	      );
      my $char = at($x + $try->[0], $y + $try->[1]);
      if ($char and ($char eq $try->[2])) {
	push(@result, trace($x + $try->[0], $y + $try->[1], [$try->[0], $try->[1]]));
      }
    }
  }
  else {
    my $char = at($x + $dir->[0], $y + $dir->[1]);
    if ($char eq ' ' or (not $char)) {
      # nothing
    }
    elsif ($char eq 'x') {
      $x += $dir->[0];
      $y += $dir->[1];
      push(@result, $inlets{$x . '_' . $y});
    }
    elsif ($char eq '*') {
      $x += $dir->[0];
      $y += $dir->[1];
      push(@result, trace($x, $y, '*', $dir));
    }
    else {
      $x += $dir->[0];
      $y += $dir->[1];
      push(@result, trace($x, $y, $dir));
    }
  }
  return(@result);
}

sub at {
  my ($x, $y) = @_;
  my $line = $lines[$y];
  return '' unless $line;
  return '' unless length($line) > $x;
  my $result = substr($line, $x, 1);

  return($result);
}
